"""Adapted from Solaris: https://github.com/CosmiQ/solaris/tree/main/solaris """

import logging
import json
import pandas as pd
import geopandas as gpd
import rasterio
import shapely
from rasterio import features
from shapely.geometry import shape

from ..config.logging_config import logger
from .geo import rasterio_load
from .geo_transforms import geojson_to_px_gdf, df_to_coco_annos, make_coco_image_dict

logger = logging.getLogger(__name__)

def mask_to_poly_geojson(mask_path,
                         output_path="", 
                         min_area=40,
                         simplify=False,
                         tolerance=1.):
    """Get polygons from an image mask.

    Args:
        mask_path : str
            Generated mask path.
        
        output_path : str, optional
            Path to save the output file to. If not provided, no file is saved.
        min_area : int, optional
            The minimum area of a polygon to retain. Filtering is done AFTER
            any coordinate transformation, and therefore will be in destination
            units.
        simplify : bool, optional
            If ``True``, will use the Douglas-Peucker algorithm to simplify edges,
            saving memory and processing time later. Defaults to ``False``.
        tolerance : float, optional
            The tolerance value to use for simplification with the Douglas-Peucker
            algorithm. Defaults to ``0.5``. Only has an effect if
            ``simplify=True``.

    Returns:
        gdf : :class:`geopandas.GeoDataFrame`
            A GeoDataFrame of polygons.

    """
    
    with rasterio.open(mask_path, 'r') as src:
        raster = src.read()
        transform = src.transform
        crs = src.crs
    mask = raster > 0
    polygon_generator = features.shapes(raster, transform=transform, mask=mask)
    polygons = []
    values = []  # pixel values for the polygon in mask_arr
    for polygon, value in polygon_generator:
        p = shape(polygon).buffer(0.0)
        if p.area >= min_area:
            polygons.append(shape(polygon).buffer(0.0))
            values.append(value)

    polygon_gdf = gpd.GeoDataFrame({"geometry": polygons, "value": values}, 
                                   crs=crs.to_wkt())
    if simplify:
        polygon_gdf['geometry'] = polygon_gdf['geometry'].apply(lambda x: x.simplify(tolerance=tolerance))
    
    polygon_gdf.to_file(output_path, driver='GeoJSON')
    logger.info(f"GeoJSON saved to {output_path}")

def gdf_to_yolo(geojson_path="", mask_path="", output_path="", column='value',
                im_size=(0, 0), min_overlap=0.66):
    """Convert a geodataframe containing polygons to yolo/yolt format.

    Args:
        geodataframe : str
            Path to a :class:`geopandas.GeoDataFrame` with a column named
            ``'geometry'``.  Can be created from a geojson with labels for unique
            objects. Can be converted to this format with
            ``geodataframe=gpd.read_file("./xView_30.geojson")``.
        im_path : str
            Path to a georeferenced image (ie a GeoTIFF or png created with GDAL)
            that geolocates to the same geography as the `geojson`(s). If a
            directory, the bounds of each GeoTIFF will be loaded in and all
            overlapping geometries will be transformed. This function will also
            accept a :class:`osgeo.gdal.Dataset` or :class:`rasterio.DatasetReader`
            with georeferencing information in this argument.
        output_dir : str
            Path to an output directory where all of the yolo readable text files
            will be placed.
        column : str, optional
            The column name that contains a unique integer id for each of object
            class.
        im_size : tuple, optional
            A tuple specifying the x and y heighth of a an image.  If specified as
            ``(0,0)`` (the default,) then the size is determined automatically.
        min_overlap : float, optional
            A float value ranging from 0 to 1.  This is a percantage.  If a polygon
            does not overlap the image by at least min_overlap, the polygon is
            discarded.  i.e. 0.66 = 66%. Default value of 0.66.

    Returns:
        gdf : :class:`geopandas.GeoDataFrame`.
            The csv file will be written to the output_dir.
    """

    if im_size == (0, 0):
        src = rasterio_load(mask_path)
        im_size = (src.width, src.height)
            
    gdf = gpd.read_file(geojson_path)

    [x0, y0, x1, y1] = [0, 0, im_size[0], im_size[1]]
    out_coords = [[x0, y0], [x0, y1], [x1, y1], [x1, y0]]
    points = [shapely.geometry.Point(coord) for coord in out_coords]
    pix_poly = shapely.geometry.Polygon([[p.x, p.y] for p in points])
    dw = 1. / im_size[0]
    dh = 1. / im_size[1]
    header = [column, "x", "y", "w", "h"]
    output = output_path
    gdf = geojson_to_px_gdf(gdf, mask_path)
    gdf['area'] = gdf['geometry'].area
    gdf['intersection'] = (
        gdf['geometry'].intersection(pix_poly).area / gdf['area'])
    gdf = gdf[gdf['area'] != 0]
    gdf = gdf[gdf['intersection'] >= min_overlap]
    if not gdf.empty:
        boxy = gdf['geometry'].bounds
        boxy['xmid'] = (boxy['minx'] + boxy['maxx']) / 2.0
        boxy['ymid'] = (boxy['miny'] + boxy['maxy']) / 2.0
        boxy['w0'] = (boxy['maxx'] - boxy['minx'])
        boxy['h0'] = (boxy['maxy'] - boxy['miny'])
        boxy['x'] = boxy['xmid'] * dw
        boxy['y'] = boxy['ymid'] * dh
        boxy['w'] = boxy['w0'] * dw
        boxy['h'] = boxy['h0'] * dh
        if not boxy.empty:
            gdf = gdf.join(boxy)
            gdf.to_csv(path_or_buf=output, sep=' ', columns=header, index=False, header=False)
            logger.info(f"Yolo file saved to {output}")

def geojson2coco(image_src, label_src, output_path=None, category_attribute="value", score_attribute=None,
                 preset_categories=None, include_other=True, info_dict=None,
                 license_dict=None, verbose=0):
    """Generate COCO-formatted labels from mask and polygon geojson.

    Args:
        image_src (str): A string path to an image.
        label_src (str): A string path to a geojson.
        output_path (str, optional): The path to save the JSON-formatted COCO records to. If not provided,
            the records will only be returned as a dict, and not saved to file.
        category_attribute (str, optional): The name of an attribute in the geojson that specifies which category
            a given instance corresponds to. If not provided, it's assumed that only one class of object is present
            in the dataset, which will be termed "other" in the output json.
        score_attribute (str, optional): The name of an attribute in the geojson that specifies the prediction
            confidence of a model.
        preset_categories (list of dicts, optional): A pre-set list of categories to use for labels. These categories
            should be formatted per the COCO category specification. Example:
            [{'id': 1, 'name': 'Fighter Jet', 'supercategory': 'plane'},
            {'id': 2, 'name': 'Military Bomber', 'supercategory': 'plane'}, ... ]
        include_other (bool, optional): If set to True, and preset_categories is provided, objects that don't fall
            into the specified categories will not be removed from the dataset. They will instead be passed into a
            category named "other" with its own associated category id. If False, objects whose categories don't match
            a category from preset_categories will be dropped.
        info_dict (dict, optional): A dictionary with the following key-value pairs:
            - "year": int year of creation
            - "version": str version of the dataset
            - "description": str string description of the dataset
            - "contributor": str who contributed the dataset
            - "url": str URL where the dataset can be found
            - "date_created": datetime.datetime when the dataset was created
        license_dict (dict, optional): A dictionary containing the licensing information for the dataset, with the
            following key-value pairs:
            - "name": str the name of the license.
            - "url": str a link to the dataset's license.
            Note: This implementation assumes that all of the data uses one license. If multiple licenses are provided,
            the image records will not be assigned a license ID.
        verbose (int, optional): Verbose text output. By default, none is provided; if True or 1, information-level
            outputs are provided; if 2, extremely verbose text is output.

    Returns:
        dict: A dictionary following the COCO dataset specification. Depending on arguments provided, it may or may
        not include license and info metadata.
    """
    logger.debug('Loading labels.')
    label_df = pd.DataFrame({'label_fname': [],
                             'category_str': [],
                             'geometry': []})
    curr_gdf = gpd.read_file(label_src)

    curr_gdf['label_fname'] = label_src
    curr_gdf['image_fname'] = ''
    curr_gdf['image_id'] = 1
    if category_attribute is None:
        logger.debug('No category attribute provided. Creating a default "other" category.')
        curr_gdf['category_str'] = 'other'  # add arbitrary value
        tmp_category_attribute = 'category_str'
    else:
        tmp_category_attribute = category_attribute
   
    logger.debug('Converting to pixel coordinates.')
    curr_gdf = geojson_to_px_gdf(curr_gdf, image_src)
    curr_gdf = curr_gdf.rename(columns={tmp_category_attribute: 'category_str'})
    if score_attribute is not None:
        curr_gdf = curr_gdf[['image_id', 'label_fname', 'category_str', score_attribute, 'geometry']]
    else:
        curr_gdf = curr_gdf[['image_id', 'label_fname', 'category_str', 'geometry']]
    label_df = pd.concat([label_df, curr_gdf], axis='index', ignore_index=True, sort=False)

    logger.info('Generating COCO-formatted annotations.')
    coco_dataset = df_to_coco_annos(label_df,
                                    geom_col='geometry',
                                    image_id_col='image_id',
                                    category_col='category_str',
                                    score_col=score_attribute,
                                    preset_categories=preset_categories,
                                    include_other=include_other,
                                    verbose=verbose)

    logger.debug('Generating COCO-formatted image and license records.')
    if license_dict is not None:
        logger.debug('Getting license ID.')
        if len(license_dict) == 1:
            logger.debug('Only one license present; assuming it applies to all images.')
            license_id = 1
        else:
            logger.debug('Zero or multiple licenses present. Not trying to match to images.')
            license_id = None
        logger.debug('Adding licenses to dataset.')
        coco_licenses = []
        license_idx = 1
        for license_name, license_url in license_dict.items():
            coco_licenses.append({'name': license_name,
                                  'url': license_url,
                                  'id': license_idx})
            license_idx += 1
        coco_dataset['licenses'] = coco_licenses
    else:
        logger.debug('No license information provided, skipping for image COCO records.')
        license_id = None
    coco_image_records = make_coco_image_dict({image_src: 1}, license_id)
    coco_dataset['images'] = coco_image_records

    if info_dict is not None:
        coco_dataset['info'] = info_dict

    
    with open(output_path, 'w') as outfile:
        json.dump(coco_dataset, outfile)
    logger.info(f"CocoJson file saved to {output_path}")





import logging.config
from datetime import datetime
from pathlib import Path

import yaml

script_dir = Path(__file__).resolve().parent
CONFIG_DIR = script_dir / "log_config.yaml"
USER_CACHE = Path.home().joinpath(".cache")
LOG_DIR = LOG_DIR = USER_CACHE.joinpath("geo-inference/logs")
LOG_DIR.mkdir(parents=True, exist_ok=True)

timestamp = datetime.now().strftime("%Y%m%d-%H_%M_%S")
logfilename =  f"{LOG_DIR}/{timestamp}.log"

with open(CONFIG_DIR, "r") as f:
    config = yaml.safe_load(f.read())
    config["handlers"]["file"]["filename"] = logfilename    
    logging.config.dictConfig(config)

logger = logging.getLogger(__name__)


from pathlib import Path
from typing import Any, Dict

import os
import pytest
import numpy as np
import scipy.signal.windows as w
import rasterio as rio
import torch
from geo_inference.geo_blocks import RasterDataset, InferenceSampler, InferenceMerge


@pytest.fixture
def test_data_dir():
    return Path(__file__).parent / "data"

@pytest.fixture
def raster_dataset(test_data_dir):
    image_asset = str(test_data_dir / "0.tif")
    return RasterDataset(image_asset)

class TestRasterDataset:
    def test_init(self, raster_dataset):
        assert isinstance(raster_dataset.src, rio.DatasetReader)
        assert raster_dataset.bands > 0
        assert raster_dataset.res > 0
        assert raster_dataset._crs is not None

    def test_getitem(self, raster_dataset):
        query: Dict[str, Any] = {
            'path': raster_dataset.image_asset,
            'window': (0, 0, 10, 10),  # replace with actual window
            'pixel_coords': (0, 0, 10, 10)  # replace with actual pixel_coords
        }
        sample = raster_dataset.__getitem__(query)
        assert isinstance(sample, dict)
        assert 'image' in sample
        assert 'crs' in sample
        assert 'pixel_coords' in sample
        assert 'window' in sample
        assert 'path' in sample

    def test_get_tensor(self, raster_dataset):
        query = (0, 0, 10, 10)  # replace with actual query
        size = 10  # replace with actual size
        tensor = raster_dataset._get_tensor(query, size)
        assert isinstance(tensor, torch.Tensor)
        assert tensor.shape[-2:] == (size, size)

    def test_pad_patch(self):
        x = torch.rand((3, 5, 5))
        patch_size = 10
        padded = RasterDataset.pad_patch(x, patch_size)
        assert isinstance(padded, torch.Tensor)
        assert padded.shape[-2:] == (patch_size, patch_size)


class TestInferenceSampler:
    @pytest.fixture
    def inference_sampler(self, raster_dataset):
        size = (10, 10)
        stride = (5, 5)
        return InferenceSampler(raster_dataset, size, stride)

    def test_init(self, inference_sampler):
        assert inference_sampler.size == (10, 10)
        assert inference_sampler.stride == (5, 5)
        assert inference_sampler.length > 0

    def test_iter(self, inference_sampler):
        for sample in inference_sampler:
            assert isinstance(sample, dict)
            assert 'pixel_coords' in sample
            assert 'path' in sample
            assert 'window' in sample

    def test_len(self, inference_sampler):
        assert len(inference_sampler) == inference_sampler.length

    def test_generate_corner_windows(self):
        window_size = 10
        step = window_size >> 1
        windows = InferenceSampler.generate_corner_windows(window_size)
        center_window = np.matrix(w.hann(M=window_size, sym=False))
        center_window = center_window.T.dot(center_window)
        window_top = np.vstack([np.tile(center_window[step:step + 1, :], (step, 1)), center_window[step:, :]])
        window_bottom = np.vstack([center_window[:step, :], np.tile(center_window[step:step + 1, :], (step, 1))])
        window_left = np.hstack([np.tile(center_window[:, step:step + 1], (1, step)), center_window[:, step:]])
        window_right = np.hstack([center_window[:, :step], np.tile(center_window[:, step:step + 1], (1, step))])
        window_top_left = np.block([[np.ones((step, step)), window_top[:step, step:]],
                                    [window_left[step:, :step], window_left[step:, step:]]])
        window_top_right = np.block([[window_top[:step, :step], np.ones((step, step))],
                                     [window_right[step:, :step], window_right[step:, step:]]])
        window_bottom_left = np.block([[window_left[:step, :step], window_left[:step, step:]],
                                       [np.ones((step, step)), window_bottom[step:, step:]]])
        window_bottom_right = np.block([[window_right[:step, :step], window_right[:step, step:]],
                                        [window_bottom[step:, :step], np.ones((step, step))]])
        assert isinstance(windows, np.ndarray)
        assert windows.shape == (3, 3, window_size, window_size)
        assert np.all(windows >= 0) and np.all(windows <= 1)
        assert np.allclose(windows[1, 1], center_window)
        assert np.allclose(windows[0, 1], window_top)
        assert np.allclose(windows[2, 1], window_bottom)
        assert np.allclose(windows[1, 0], window_left)
        assert np.allclose(windows[1, 2], window_right)
        assert np.allclose(windows[0, 0], window_top_left)
        assert np.allclose(windows[0, 2], window_top_right)
        assert np.allclose(windows[2, 0], window_bottom_left)
        assert np.allclose(windows[2, 2], window_bottom_right)

class TestInferenceMerge:
    @pytest.fixture
    def inference_merge(self):
        height = 100
        width = 100
        classes = 3
        device = torch.device('cpu')
        return InferenceMerge(height, width, classes, device)

    def test_init(self, inference_merge):
        assert inference_merge.height == 100
        assert inference_merge.width == 100
        assert inference_merge.classes == 3
        assert inference_merge.device == torch.device('cpu')
        assert isinstance(inference_merge.image, np.ndarray)
        assert isinstance(inference_merge.norm_mask, np.ndarray)

    def test_merge_on_cpu(self, inference_merge):
        batch = torch.rand((3, 3, 10, 10))
        windows = torch.rand((3, 1, 10, 10))
        pixel_coords = [(0, 0, 10, 10), (10, 10, 10, 10), (20, 20, 10, 10)]
        inference_merge.merge_on_cpu(batch, windows, pixel_coords)
        assert np.all(inference_merge.image >= 0)
        assert np.all(inference_merge.norm_mask >= 1)

    def test_save_as_tiff(self, inference_merge, test_data_dir):
        height = 100
        width = 100
        output_meta = {
            "crs": "+proj=latlong",
            "transform": rio.Affine(1.0, 0, 0, 0, 1.0, 0)
        }
        output_path = test_data_dir / "test_1.tiff"
        inference_merge.save_as_tiff(height, width, output_meta, output_path)
        assert output_path.exists()
        os.remove(output_path)